import logging
import zipfile
from pathlib import Path
from typing import Dict, Sequence, Tuple, Union

import numpy as np
from osgeo import gdal, osr

from . import utils
from .base import Image

logger = logging.getLogger(__name__)

__all__ = ["load_image", "load_from_zip"]


def load_image(
    img_path: Union[Path, str],
    bbox: Tuple = None,
    fill_outside: bool = False,
    masked: bool = False,
    **kwargs,
) -> Image:
    """Load a raster file into an :class:`~resens.base.Image`.

    The function reads the array, geo-transform, projection and metadata using GDAL.

    :param img_path: Path to image
    :param bbox: Tuple containing bounding box coordinatee to be used for clipping the
        image. Should be formatted as (xmin, ymin, xmax, ymax).
    :param fill_outside: If ``True``, pixels outside raster coverage (when ``bbox`` exceed
        the raster extent) are padded with zeros.
    :param masked: If ``True`` and the raster has an assigned nodata value, pixels with
        this value will be converted to `nan`. Masking the array will convert it to
        float32, increasing the memory consumption.
    :param kwargs: Optional overrides for ``transformation`` and ``projection``.
    :return: Loaded :class:`~resens.base.Image`.
    """
    load_kwargs = {}

    img_path = img_path.as_posix() if isinstance(img_path, Path) else img_path
    dataset: gdal.Dataset = gdal.Open(img_path)

    # Get geographic information
    transf = kwargs.get("transformation", None) or list(dataset.GetGeoTransform())
    proj = (
        kwargs.get("projection", None)
        or dataset.GetProjection()
        or dataset.GetGCPProjection()
    )

    srs = osr.SpatialReference(wkt=proj)
    srs = osr.SpatialReference()
    srs.ImportFromWkt(proj)
    srs.AutoIdentifyEPSG()
    epsg = (
        srs.GetAuthorityCode(None)
        or srs.GetAuthorityCode("PROJCS")
        or srs.GetAuthorityCode("GEOGCS")
    )

    # in case the raster is geoereferenced with GCPs, warp it to create a raster with
    # geotransformation
    if transf == [0, 1, 0, 0, 0, 1] and len(dataset.GetGCPs()):
        dataset = gdal.Warp("/vsimem/warp.tif", dataset, dstSRS=f"EPSG:{epsg}")
        transf = list(dataset.GetGeoTransform())

    # Get metadata
    metadata = dataset.GetMetadata()
    nodata_val = dataset.GetRasterBand(1).GetNoDataValue()

    # If bounds have been passed, calculate the extents for clipping
    if bbox:
        raster_size = dataset.RasterXSize, dataset.RasterYSize
        xmin, ymin, xmax, ymax = bbox
        xo, px, _, yo, _, py = transf
        xoff = int(np.ceil((xmin - xo) / px))
        yoff = int(np.ceil((ymax - yo) / py))
        padx = [0, 0]
        pady = [0, 0]
        if fill_outside:
            if xoff < 0:
                padx[0] = abs(xoff)
            if yoff < 0:
                pady[0] = abs(xoff)
        xoff = 0 if xoff < 0 else xoff
        yoff = 0 if yoff < 0 else yoff

        xsize = int(np.floor((xmax - xo) / px)) - xoff
        ysize = int(np.floor((ymin - yo) / py)) - yoff
        if fill_outside:
            if xsize + xoff > raster_size[0]:
                padx[1] = xsize + xoff - raster_size[0]
                xsize -= padx[1]
            if ysize + yoff > raster_size[1]:
                pady[1] = ysize + yoff - raster_size[1]
                ysize -= pady[1]

        if xsize < 0 or ysize < 0:
            if px == py:
                raise ValueError(
                    "Negative dimensions encountered when cropping. "
                    "Image transformation is invalid."
                )
            else:
                raise ValueError(
                    "Negative dimensions encountered when cropping. "
                    "Check bound coordinates."
                )

        load_kwargs["xoff"] = xoff
        load_kwargs["yoff"] = yoff
        load_kwargs["xsize"] = xsize
        load_kwargs["ysize"] = ysize

    # Read array
    array: np.ndarray = dataset.ReadAsArray(**load_kwargs)
    if array.ndim == 3:
        array = np.einsum("ijk->jki", array)

    if nodata_val is not None and masked is True:
        array = array.astype(np.float32)
        array[array == nodata_val] = np.nan
        dtype_s = "float32"
    else:
        dtype_s, dtype_np = utils.find_dtype(array)
        array = array.astype(dtype_np)

    # Pad the array if the selected bounds are outside of its dimensions
    if bbox and fill_outside:
        pad_widths = [pady, padx]

        if array.ndim == 3:
            pad_widths += [[0, 0]]  # 2-axis padding (no padding)
        array = np.pad(array, pad_widths, "constant")

    # Check that the pixel sizes are of the correct sign
    xo, psx, skx, yo, sky, psy = list(transf)
    if psy > 0:
        psy = -psy
    if bbox:
        xo += (xoff - padx[0]) * psx
        yo += (yoff - pady[0]) * psy
    transf = [xo, psx, skx, yo, sky, psy]

    dataset = None

    return Image(array, transf, proj, epsg, metadata, nodata_val, dtype_s)


def load_from_zip(
    zipf_path: Union[Path, str],
    patterns: Union[str, Sequence[str]],
    extension: str,
    group: str = "",
    bounds: Tuple = None,
    fill_outside: bool = False,
    masked: bool = False,
) -> Union[Dict[str, Image], None]:
    """Load rasters from within a ZIP archive.

    Files are selected by:
    - matching the provided file ``extension``,
    - containing any of the strings in ``patterns``,
    - and containing ``group`` (if provided).

    :param zipf_path: Path to zip file
    :param patterns: List of strings included in the file names (e.g. band numbers)
    :param extension: Extension of the target image
    :param group: Extra string to search for.
    :param bounds: Tuple containing bounds to be used for clipping the image.
        Should be formatted as ((xmin, xmax), (xmax, ymin)).
    :param fill_outside: If ``True``, pad outside coverage with zeros when cropping.
    :param masked: If ``True`` and the raster has an assigned nodata value, pixels with
        this value will be converted to `nan`. Masking the array will convert it to
        float32, increasing the memory consumption.
    :return: A dictionary mapping each matched key from ``patterns`` to an
        :class:`~resens.base.Image`. Returns ``None`` if the ZIP cannot be read.
    """

    # ensure ``patterns`` is a list
    if not isinstance(patterns, (list, tuple)):
        patterns = [patterns]

    # Check if the zip file path is correct
    if isinstance(zipf_path, str):
        zipf_path = Path(zipf_path)
    if not zipf_path.exists():
        raise FileNotFoundError(f"Zip file {zipf_path} does not exist!")

    # Initialize gdal zip file handler
    ziphandler = "/vsizip/"

    # Read zip file
    try:
        archive = zipfile.ZipFile(zipf_path, "r")
    except zipfile.BadZipfile:
        return None
    else:
        # Get the zip structure for the required bands
        img_ls = [
            f
            for f in archive.namelist()
            if f.endswith(extension) and any(x in f for x in patterns) and group in f
        ]

        # Create dictionaries to store the data
        band_dict = {}

        for img in img_ls:
            try:
                # Find which of the req files fits the current, create the dict key
                # and store it in the keys' list
                (key_in,) = [key for key in patterns if key in img]

                if key_in in band_dict:
                    logger.warning(f"Multiple files were found for key: {key_in}!")
                    continue

                # Load image, get metadata and store to dictionary
                band_dict[key_in] = load_image(
                    img_path=ziphandler + zipf_path.joinpath(img).as_posix(),
                    bbox=bounds,
                    fill_outside=fill_outside,
                    masked=masked,
                )
            except AttributeError:
                raise AttributeError(f"Error loading {img}")

        return band_dict



import tempfile
import unittest
from pathlib import Path

import numpy as np

import resens as rs
from resens import utils

base_path = Path(__file__).parent


class TestSum(unittest.TestCase):
    def test_load_image(self):
        image = rs.load_image(base_path.joinpath("data", "sample-bgrn-16bit-small.tif"))
        self.assertTupleEqual(
            image.array.shape, (1511, 1441, 4), "Array has the correct dimensions"
        )
        self.assertEqual(image.epsg_code, "32639", "The correct EPSG code was loaded")

    def test_write_image(self):
        # First load the sample image
        sample = rs.load_image(base_path.joinpath("data", "sample-bgrn-16bit-small.tif"))

        # Then write a test output image
        output_path = Path(tempfile.gettempdir(), "test_output.tif")
        sample.write_image(
            path=output_path,
            compression=True,
        )

        # Then load the test output
        image = rs.load_image(output_path)

        # Now check to make sure everything is correct
        self.assertTrue(np.all(sample.array == image.array), "Arrays are not equal")
        self.assertListEqual(
            sample.transformation, image.transformation, "Transformation is correct"
        )
        self.assertEqual(sample.projection, image.projection, "Projection is correct")
        self.assertEqual(sample.epsg_code, image.epsg_code, "EPSG code is correct")
        self.assertDictEqual(sample.metadata, image.metadata, "Metadata is correct")

    def test_get_sliding_win(self):

        # initialize two random arrays
        arr_sb = np.random.randint(0, 256, (100, 100))
        arr_mb = np.random.randint(0, 256, (100, 100, 3))

        arr_sb_convs = utils.get_sliding_win(
            array=arr_sb, ksize=3, step_x=1, step_y=1, pad=True
        )
        self.assertTupleEqual(
            arr_sb_convs.shape,
            (102, 102, 3, 3),
            "Correct convolution number (singleband)",
        )

        arr_mb_convs = utils.get_sliding_win(
            array=arr_mb, ksize=3, step_x=1, step_y=1, pad=True
        )
        self.assertTupleEqual(
            arr_mb_convs.shape,
            (102, 102, 3, 3, 3),
            "Correct tile number (multiband)",
        )

    def test_get_tiles(self):

        # initialize two random arrays
        arr_sb = np.random.randint(0, 256, (100, 100))
        arr_mb = np.random.randint(0, 256, (100, 100, 3))

        arr_sb_tiles = utils.get_tiles(
            array=arr_sb,
            ksize=3,
        )
        self.assertTupleEqual(
            arr_sb_tiles.shape, (34, 34, 3, 3), "Correct tile number (singleband)"
        )

        arr_mb_tiles = utils.get_tiles(
            array=arr_mb,
            ksize=3,
        )
        self.assertTupleEqual(
            arr_mb_tiles.shape, (34, 34, 3, 3, 3), "Correct tile number (multiband)"
        )


if __name__ == "__main__":
    unittest.main()



import logging
from dataclasses import dataclass, field
from numbers import Number
from pathlib import Path
from typing import Dict, List, Literal, Optional, Sequence, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.image import AxesImage
from osgeo import gdal, osr
from pyproj import Transformer

from . import rasteroptions, utils

logger = logging.getLogger(__name__)

__all__ = ["Image"]


@dataclass
class _Image:
    array: np.ndarray = field(default_factory=lambda: np.ndarray([]))
    transformation: List = field(default_factory=lambda: [0, 1, 0, 0, 0, -1])
    projection: str = ""
    epsg_code: int = 1
    metadata: Dict = field(default_factory=dict)
    nodata: Number = None
    dtype: str = None


class Image(_Image):
    def copy(self) -> "Image":
        """Creates a deep copy of the :class:`~resens.base.Image` instance.

        :return: Returns a copy of :class:`~resens.base.Image`
        """
        return Image(
            array=self.array.copy(),
            transformation=self.transformation.copy(),
            projection=self.projection,
            epsg_code=self.epsg_code,
            metadata=self.metadata.copy(),
        )

    def extents(
        self, dst_epsg: Optional[int] = None
    ) -> Tuple[float, float, float, float]:
        """Return the bounding box of a georeferenced image.

        The bounds are derived from the image geo-transform and array shape.
        :param dst_epsg: Pass an EPSG code to transform to another CRS, defaults to None.

        :return: Bounding box as ``(xmin, ymin, xmax, ymax)``.
        """
        shape = self.array.shape[:2]
        xmin, psx, _, ymax, _, psy = self.transformation

        bbox = (
            xmin,
            ymax + shape[0] * psy + psy / 2,
            xmin + shape[1] * psx + psx / 2,
            ymax,
        )

        # optional: transform extents bbox to another CRS using an EPSG code
        if dst_epsg is not None and dst_epsg != self.epsg_code:
            transformer = Transformer.from_crs(
                f"EPSG:{self.epsg_code}",
                f"EPSG:{dst_epsg}",
                always_xy=True,  # ensures consistent (lon, lat) / (x, y) order
            )

            min_x, min_y = transformer.transform(bbox[0], bbox[1])
            max_x, max_y = transformer.transform(bbox[2], bbox[3])

            bbox = (min_x, min_y, max_x, max_y)

        return bbox

    def resample(
        self,
        out_shape: Optional[Sequence[int]] = None,
        out_pix: Optional[Number] = None,
        interpolation: Literal[
            "near",
            "bilinear",
            "cubic",
            "cubicspline",
            "lanczos",
            "average",
            "rms",
            "mode",
            "max",
            "min",
            "med",
            "q1",
            "q3",
            "sum",
        ] = "bilinear",
        inplace: bool = False,
    ) -> Optional["Image"]:
        """Resample the image array by output shape or output pixel size.

        Exactly one of ``out_shape`` or ``out_pix`` should be provided.

        :param out_shape: Output array shape as ``(nrows, ncols)``.
        :param out_pix: Output pixel size. For square pixels pass a single number.
                        For rectangular pixels pass ``(psx, psy)``.
        :param interpolation: Interpolation method: ``near``, ``bilinear``, ``cubic``,
                              ``cubicspline``, ``lanczos``, ``average``, ``rms``,
                              ``mode``, ``max``, ``min``, ``med``, ``q1``, ``q3``,
                              ``sum``
        :param inplace: If ``True``, modify the current object and return ``None``.
                        If ``False``, return a resampled copy.
        :return: A new :class:`~resens.base.Image` or ``None`` if ``inplace=True``.
        """
        src_ds = self.write_image("MEM")

        # Warp to in-memory output
        if out_shape is not None:
            resampled_ds: gdal.Dataset = gdal.Warp(
                "",
                src_ds,
                format="MEM",
                dstSRS=f"EPSG:{self.epsg_code}",
                width=out_shape[1],
                height=out_shape[0],
                resampleAlg=interpolation,
            )
        elif out_pix is not None:
            if not isinstance(out_pix, list):
                out_pix = [out_pix, out_pix]
            else:
                if not len(out_pix) == 2:
                    out_pix *= 2

            resampled_ds: gdal.Dataset = gdal.Warp(
                "",
                src_ds,
                format="MEM",
                dstSRS=f"EPSG:{self.epsg_code}",
                xRes=out_pix[0],
                yRes=out_pix[1],
                resampleAlg=interpolation,
            )

        if resampled_ds is None:
            raise RuntimeError("Failed to resample image!")

        # Read reprojected data
        resampled = self if inplace else self.copy()
        resampled.array = resampled_ds.ReadAsArray()
        if resampled.array.ndim == 3:
            resampled.array = np.einsum("ijk->jki", resampled.array)
        resampled.array = resampled.array.astype(utils.find_dtype(resampled.array)[1])

        resampled.transformation = list(resampled_ds.GetGeoTransform())
        resampled.projection = resampled_ds.GetProjection()
        resampled.epsg_code = int(
            osr.SpatialReference(wkt=resampled.projection).GetAttrValue("AUTHORITY", 1)
        )
        resampled.metadata = resampled_ds.GetMetadata()

        # Clean up
        src_ds = None
        resampled_ds = None

        return resampled if inplace is False else None

    def reproject(self, target_epsg: int, inplace: bool = False) -> Optional["Image"]:
        """Reproject the image to a target CRS.

        :param target_epsg: Target EPSG code. If it matches the source EPSG, the result
                            will be equivalent to the input.
        :param inplace: If ``True``, modify the current object and return ``None``.
                        If ``False``, return a reprojected copy.
        :return: A new :class:`~resens.base.Image` or ``None`` if ``inplace=True``.
        """
        src_ds = self.write_image("MEM")

        # Warp to in-memory output
        reproj_ds: gdal.Dataset = gdal.Warp(
            "", src_ds, format="MEM", dstSRS=f"EPSG:{target_epsg}"
        )

        if reproj_ds is None:
            raise RuntimeError(f"Failed to reproject image to EPSG:{target_epsg}")

        # Read reprojected data
        reproj = self if inplace else self.copy()
        reproj.array = reproj_ds.ReadAsArray()
        if reproj.array.ndim == 3:
            reproj.array = np.einsum("ijk->jki", reproj.array)
        reproj.array = reproj.array.astype(utils.find_dtype(reproj.array)[1])

        reproj.transformation = list(reproj_ds.GetGeoTransform())
        reproj.projection = reproj_ds.GetProjection()
        reproj.epsg_code = int(
            osr.SpatialReference(wkt=reproj.projection).GetAttrValue("AUTHORITY", 1)
        )
        reproj.metadata = reproj_ds.GetMetadata()

        # Clean up
        src_ds = None
        reproj_ds = None

        return reproj if inplace is False else None

    def to_8bit(self, inplace: bool = False) -> Optional["Image"]:
        """Convert the image to 8-bit with simple contrast enhancement.

        The conversion uses histogram truncation based on 2nd-98th percentiles.

        :param inplace: If ``True``, modify the current object and return ``None``.
                        If ``False``, return a converted copy.
        :return: A new :class:`~resens.base.Image` or ``None`` if ``inplace=True``.
        """

        def _make_8bit(arr: np.ndarray) -> np.ndarray:
            # Get image statistics
            min_val = np.nanpercentile(arr, 2)
            min_val = min_val if min_val >= 0 else 0
            max_val = np.nanpercentile(arr, 98)

            # Truncate the array - Contrast Enhancement
            arr = np.clip(arr, min_val, max_val)

            # Convert to 8bits
            arr = np.divide(arr - min_val, max_val - min_val) * 255

            return arr.astype(np.uint8)

        conv_8bit = self if inplace else self.copy()
        bit8_img = conv_8bit.array.astype(np.float32)

        if bit8_img.ndim == 3:
            for i in range(bit8_img.shape[2]):
                bit8_img[:, :, i] = _make_8bit(bit8_img[:, :, i])
        else:
            bit8_img = _make_8bit(bit8_img)

        conv_8bit.array = bit8_img
        return conv_8bit if not inplace else None

    def to_grayscale(self, inplace: bool = False) -> Optional["Image"]:
        """
        Method to convert a multiband Image to single band (grayscale).

        :param inplace: If ``True``, modify the current object and return ``None``.
                        If ``False``, return a grayscale copy.
        :return: A new :class:`~resens.base.Image` or ``None`` if ``inplace=True``.
        """
        # If the image is not composed of multiple channels, it is already grayscale
        try:
            assert self.array.ndim == 3
        except AssertionError:
            return self.array

        grayscale = self if inplace else self.copy()
        img_sum = np.sum(self.array, axis=2)  # Sum of all channel pixels

        # Compute weight for each channel
        img_weights_arr = np.dstack(
            [np.divide(self.array[..., i], img_sum) for i in range(self.array.shape[2])]
        )
        channel_weights = np.nanmean(img_weights_arr, axis=(0, 1))
        while channel_weights.sum() > 1.0:
            channel_weights -= channel_weights * 0.01  # Make sure the weights sum to 1.0

        # Get grayscale image
        grayscale.array = np.nansum(self.array * channel_weights, axis=2)

        return grayscale if not inplace else None

    def write_image(
        self,
        path: Union[Path, str, Literal["MEM"]],
        nodata: Number = None,
        compression: bool = True,
        datatype: str = None,
        metadata: Dict = None,
    ) -> Union[gdal.Dataset, None]:
        """Write the image to a GeoTIFF on disk.

        :param path: Output path. If no suffix is provided, ``.tif`` is appended. Set to
            ``"MEM"`` to write the raster to memory.
        :param nodata: NoData value to set on the output band(s), if provided.
        :param compression: If ``True``, enable DEFLATE compression. Not valid if
            ``path="MEM"``.
        :param datatype: Output array datatype. If ``None``, the datatype is inferred.
            Supported: ``uint8``, ``uint16``, ``int8``, ``int16``, ``float32``.
        :param metadata: Extra metadata to write to the output dataset. If provided, it
            is merged with existing ``self.metadata``.
        :return: gdal.Dataset if ``path="MEM"``, otherwise ``None``.
        """
        if nodata is None:
            nodata = self.nodata

        # Check that the pixel sizes are of the correct sign
        if self.transformation[-1] > 0:
            self.transformation[-1] *= -1

        # Get array type
        datatype = datatype or self.dtype
        if datatype is None:
            datatype, _ = utils.find_dtype(self.array)
            self.array = self.array.astype(datatype)

        gdal_datatype = rasteroptions.GDAL_DTYPES[datatype]

        try:
            # Determine the shape of the array and the number of bands
            if self.array.shape[0] > self.array.shape[2]:
                row_ind = 0
                col_ind = 1
                nband = self.array.shape[2]
            else:
                row_ind = 1
                col_ind = 2
                nband = self.array.shape[0]

        except IndexError:
            row_ind = 0
            col_ind = 1
            nband = 1

        # Construct output image path
        if not path == "MEM":
            if isinstance(path, str):
                path = Path(path)
            if path.suffix == "":
                path = path.with_suffix(".tif")

            driver: gdal.Driver = gdal.GetDriverByName("GTiff")
        else:
            driver: gdal.Driver = gdal.GetDriverByName("MEM")
        dataset: gdal.Dataset = driver.Create(
            path.as_posix() if not path == "MEM" else "",
            self.array.shape[col_ind],
            self.array.shape[row_ind],
            nband,
            gdal_datatype,
            options=[]
            if path == "MEM"
            else rasteroptions.CO_COMPRESS
            if compression
            else rasteroptions.CO_NOCOMPRESS,
        )
        dataset.SetGeoTransform(self.transformation)
        dataset.SetProjection(self.projection)
        if metadata is not None:
            dataset.SetMetadata({**self.metadata, **metadata})

        for i in range(nband):
            if not nband == 1:
                out_band: gdal.Band = dataset.GetRasterBand(i + 1)
                if nodata:
                    out_band.SetNoDataValue(nodata)
                out_band.WriteArray(self.array[..., i])
            else:
                out_band: gdal.Band = dataset.GetRasterBand(i + 1)
                if nodata:
                    out_band.SetNoDataValue(nodata)
                out_band.WriteArray(self.array)
            out_band = None

        dataset.FlushCache()
        if path == "MEM":
            return dataset
        else:
            dataset = None

    def imshow(
        self, image_extents: Optional[List] = None, bbox: Optional[List] = None, **kwargs
    ) -> AxesImage:
        """Plots an array with matplotlib. If a 3D array is passed.

        :param image_extents: Image extents in pixels for clipping formatted as (`minx`,
            `miny`, `maxx`, `maxy`), defaults to None. Overrides ``bbox``.
        :param bbox: Image extents in CRS coordinates for clipping formatted as (`minx`,
            `miny`, `maxx`, `maxy`), defaults to None.

        :return: AxesImage object
        """
        plot_img = self.clip(image_extents, bbox, silent=True)
        strtype = utils.find_dtype(plot_img.array)[0]

        if plot_img.array.ndim == 3:
            if plot_img.array.shape[-1] < 3:
                plot_img.array = plot_img.array[..., 0]
            else:
                plot_img.array = plot_img.array[..., :3]
        else:
            plot_img.array = plot_img.array

        if strtype != "uint8":
            ret = plt.imshow(
                np.divide(
                    plot_img.array - plot_img.array.min(),
                    plot_img.array.max() - plot_img.array.min(),
                ),
                **kwargs,
            )
        else:
            ret = plt.imshow(plot_img.array, **kwargs)

        return ret

    def clip(
        self,
        image_extents: Optional[List] = None,
        bbox: Optional[List] = None,
        inplace: bool = False,
        silent: bool = False,
    ) -> Optional["Image"]:
        """Clips an array.

        :param image_extents: Image extents in pixels for clipping formatted as (`minx`,
            `miny`, `maxx`, `maxy`), defaults to None. Overrides ``bbox``.
        :param bbox: Image extents in CRS coordinates for clipping formatted as (`minx`,
            `miny`, `maxx`, `maxy`), defaults to None.

        :return: A new :class:`~resens.base.Image` or ``None`` if ``inplace=True``.
        """
        clipped = self if inplace else self.copy()

        if image_extents is not None:
            minx, miny, maxx, maxy = image_extents
        elif bbox is not None:
            g_minx, g_miny, g_maxx, g_maxy = bbox

            minx = max(
                0, (g_minx - clipped.transformation[0]) / clipped.transformation[1]
            )
            maxx = (g_maxx - clipped.transformation[0]) / clipped.transformation[1]
            miny = max(
                0, (g_maxy - clipped.transformation[3]) / clipped.transformation[-1]
            )
            maxy = (g_miny - clipped.transformation[3]) / clipped.transformation[-1]
        else:
            if not silent:
                logger.warning(
                    "``image_extents`` or ``bbox`` need to be passed to clip the image!"
                )
            return clipped

        minx, miny, maxx, maxy = list(
            map(lambda x: int(round(x)), [minx, miny, maxx, maxy])
        )

        # clip image
        clipped.array = clipped.array[miny:maxy, minx:maxx]

        # adjust transformation
        clipped.transformation[0] -= minx * clipped.transformation[1]
        clipped.transformation[3] -= maxy * clipped.transformation[-1]

        return clipped



from pathlib import Path
from typing import List

from setuptools import setup

BASE_DIR = Path(__file__).parent
long_description = (BASE_DIR / "README.md").read_text()


def read_version() -> str:
    """Function to read the package version

    :return: Package version
    :rtype: str
    """

    version_file = BASE_DIR / "resens" / "__version__.py"
    context = {}
    exec(version_file.read_text(), context)

    return context["__version__"]


def get_required() -> List[str]:
    """Function to get package dependencies

    :return: List of package dependencies
    """

    required = []
    with open(BASE_DIR / "requirements/base.txt", encoding="UTF-8") as reqs:
        for line in reqs:
            line = line.strip()
            if not line:
                continue
            if line[0] not in {"-", "#"}:
                required.append(line)

    return required


setup(
    name="resens",
    version=read_version(),
    description="Raster Processing package for Remote Sensing and Earth Observation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Nikos Argyropoulos",
    author_email="n.argiropgeo@gmail.com",
    license="MIT",
    packages=["resens"],
    package_dir={"resens": "resens"},
    python_requires=">=3.8,<3.15",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.14",
    ],
    zip_safe=False,
    install_requires=get_required(),
)



import logging
from typing import Callable, Sequence, Tuple, Type, Union

import numpy as np
from numpy import int8, short, single, uint8, ushort

logger = logging.getLogger(__name__)

__all__ = ["find_dtype", "swf", "get_sliding_win", "get_tiles"]


def find_dtype(
    in_arr: np.ndarray,
) -> Tuple[str, Type[Union[int8, uint8, short, ushort, single]]]:
    """Infer a compact dtype for an input array.

    The inference is based on whether values are integer-like and on the observed
    value range.

    :param in_arr: Input array.
    :return: Tuple ``(dtype_name, numpy_dtype)`` where ``dtype_name`` is one of
             ``{"int8", "int16", "uint8", "uint16", "float32"}``.
    """
    # Initialize a dictionary to store the numpy attribute for each dtype
    dtype_dict = {
        "int8": np.int8,
        "int16": np.int16,
        "uint8": np.uint8,
        "uint16": np.uint16,
        "float32": np.float32,
    }

    # Get minimum and maximum values in the array and the value of one random element
    min_val = np.min(in_arr)
    max_val = np.max(abs(in_arr))

    if np.array_equal(in_arr, in_arr.astype(np.int_)):
        if max_val < 256:
            if min_val < 0:
                arrtype = "int8"
                npdtype = dtype_dict[arrtype]
            else:
                arrtype = "uint8"
                npdtype = dtype_dict[arrtype]
        elif 256 <= max_val <= 65535:
            if min_val < 0:
                arrtype = "int16"
                npdtype = dtype_dict[arrtype]
            else:
                arrtype = "uint16"
                npdtype = dtype_dict[arrtype]
        else:
            arrtype = "float32"
            npdtype = dtype_dict[arrtype]
    else:
        arrtype = "float32"
        npdtype = dtype_dict[arrtype]

    return arrtype, npdtype


def swf(
    in_arr: np.ndarray, ksize: int = None, filter_op: Union[Callable, str] = "mean"
) -> np.ndarray:
    """Apply a side-window filter (mean or median) pixel-by-pixel.

    :param in_arr: Input array (2D or 3D).
    :param ksize: Window size (odd). If even, it is adjusted to the next odd number.
    :param filter_op: Filter operation: ``"median"``, ``"mean"``, or a callable.
    :return: Filtered array with the same shape and dtype as input.
    """
    from numpy.lib.stride_tricks import as_strided

    # Available filter operations
    if isinstance(filter_op, str):
        filter_dict = {"median": np.median, "mean": np.mean}
        filter_op = filter_dict[filter_op]

    # Get window radius, padded array and strides
    ksize += 1 - ksize % 2  # Make sure window size is an odd number
    radius = ksize // 2  # Window radius
    padded = np.pad(in_arr, radius, "reflect")  # Pad array using the radius
    strides = padded.strides + padded.strides

    # Parameters that depend on the input array
    try:
        assert in_arr.ndim == 2
        sy, sx = in_arr.shape
        bands = False
        reshape_shape = -1
        pr_axes = [2, 3]
    except AssertionError:
        sy, sx, bands = in_arr.shape
        reshape_shape = (1, sy * sx, bands)
        pr_axes = [3, 4]

    # Calculate output shape
    if not bands:
        up_down_shape = (sy + radius, sx, ksize - radius, ksize)
        left_right_shape = (sy, sx + radius, ksize, ksize - radius)
        others_shape = (sy + radius, sx + radius, ksize - radius, ksize - radius)
    else:
        up_down_shape = (sy + radius, sx, 1, ksize - radius, ksize, bands)
        left_right_shape = (sy, sx + radius, 1, ksize, ksize - radius, bands)
        others_shape = (
            sy + radius,
            sx + radius,
            1,
            ksize - radius,
            ksize - radius,
            bands,
        )

    # Slice the padded array using strides
    up_down = as_strided(padded, shape=up_down_shape, strides=strides)
    left_right = as_strided(padded, shape=left_right_shape, strides=strides)
    rest = as_strided(padded, shape=others_shape, strides=strides)
    padded = None

    # Get the median/mean value of each sub-window, then flatten them
    up_down_meds = np.apply_over_axes(filter_op, up_down, pr_axes).astype(up_down.dtype)
    up_down = None
    left_right_meds = np.apply_over_axes(filter_op, left_right, pr_axes).astype(
        left_right.dtype
    )
    left_right = None
    rest_meds = np.apply_over_axes(filter_op, rest, pr_axes).astype(rest.dtype)
    rest = None

    # Compute filter for subwindows
    up_meds = up_down_meds[:-radius, :].reshape(reshape_shape)
    down_meds = up_down_meds[radius:, :].reshape(reshape_shape)
    left_meds = left_right_meds[:, :-radius].reshape(reshape_shape)
    right_meds = left_right_meds[:, radius:].reshape(reshape_shape)
    nw_meds = rest_meds[:-radius, :-radius].reshape(reshape_shape)
    sw_meds = rest_meds[radius:, :-radius].reshape(reshape_shape)
    ne_meds = rest_meds[:-radius, radius:].reshape(reshape_shape)
    se_meds = rest_meds[radius:, radius:].reshape(reshape_shape)

    # Stack and find closest sub-window value per pixel
    stacked = np.vstack(
        (
            up_meds,
            down_meds,
            right_meds,
            left_meds,
            nw_meds,
            sw_meds,
            ne_meds,
            se_meds,
        )
    )

    up_meds = None
    down_meds = None
    right_meds = None
    left_meds = None
    nw_meds = None
    sw_meds = None
    ne_meds = None
    se_meds = None

    subtr = np.absolute(stacked - in_arr.reshape(reshape_shape))
    in_arr = None
    inds = np.argmin(subtr, axis=0)
    subtr = None

    if not bands:
        filt = np.take_along_axis(stacked, np.expand_dims(inds, axis=0), axis=0).reshape(
            sy, sx
        )
    else:
        filt = np.take_along_axis(stacked, np.expand_dims(inds, axis=0), axis=0).reshape(
            sy, sx, bands
        )

    stacked = None
    return filt


def get_sliding_win(
    array: np.ndarray,
    ksize: int,
    step_x: int = 1,
    step_y: int = 1,
    pad: bool = True,
) -> np.ndarray:
    """Return a strided view of sliding windows over an array.

    :param array: Input array (2D or 3D).
    :param ksize: Window size.
    :param step_x: Stride in x-direction.
    :param step_y: Stride in y-direction.
    :param pad: If ``True``, reflect-pad by ``ksize//2`` before extracting windows.
    :return: A strided array view containing windows.
    """
    from numpy.lib.stride_tricks import as_strided

    # Get window radius, padded array and strides
    if pad:
        radius = ksize // 2
        pad_widths = [
            [radius, radius],
            [radius, radius],
        ]
        if array.ndim == 3:
            pad_widths += [[0, 0]]
        array = np.pad(array, pad_widths, "reflect")
    else:
        radius = 0

    if array.ndim == 2:
        sy, sx = array.shape
        nbands = False
    elif array.ndim == 3:
        sy, sx, nbands = array.shape
    else:
        raise ValueError(f"Incorrect array shape {array.shape}")

    y_size = int(np.floor((sy + 2 * radius - ksize) / step_y) + 1)
    x_size = int(np.floor((sx + 2 * radius - ksize) / step_x) + 1)

    if not nbands:
        strides = (
            array.strides[0] * step_y,
            array.strides[1] * step_x,
        ) + array.strides
        out_shape = (y_size, x_size, ksize, ksize)
    else:
        strides = (
            array.strides[0] * step_y,
            array.strides[1] * step_x,
            array.strides[2],
        ) + array.strides
        out_shape = (y_size, x_size, 1, ksize, ksize, nbands)

    sliced_array = np.squeeze(as_strided(array, shape=out_shape, strides=strides))
    return sliced_array


def get_tiles(
    array: np.ndarray,
    ksize: Union[int, Sequence[int]] = None,
    nblocks: int = None,
    pad: bool = True,
) -> np.ndarray:
    """Split an array into non-overlapping tiles using a strided view.

    Provide either ``ksize`` (tile size) or ``nblocks`` (number of blocks).

    :param array: Input array (2D or 3D).
    :param ksize: Tile size as an int or as ``(ksize_x, ksize_y)``.
    :param nblocks: Number of tiles as an int or as ``(nblocks_x, nblocks_y)``.
    :param pad: If ``True``, reflect-pad to make the array divisible by tile size.
    :return: A strided array view containing tiles.
    """
    from numpy.lib.stride_tricks import as_strided

    if ksize:
        if isinstance(ksize, (list, tuple)):
            ksize_x = ksize[0]
            ksize_y = ksize[1]
            step_x = ksize[0]
            step_y = ksize[1]
        else:
            ksize_x = ksize
            ksize_y = ksize
            step_x = ksize
            step_y = ksize
    elif nblocks:
        if isinstance(nblocks, (list, tuple)):
            ksize_x = array.shape[1] // nblocks[0]
            ksize_y = array.shape[0] // nblocks[1]
            step_x = ksize_x
            step_y = ksize_y
        else:
            ksize_x = array.shape[1] // nblocks
            ksize_y = array.shape[0] // nblocks
            step_x = ksize_x
            step_y = ksize_y
    else:
        raise RuntimeError("One of `ksize`, `nblocks` needs to be passed!")

    # Get window radius, padded array and strides
    if pad:
        pad_widths = [
            [0, ksize - array.shape[0] % ksize],  # 0-axis padding
            [0, ksize - array.shape[1] % ksize],  # 1-axis padding
        ]
        if array.ndim == 3:
            pad_widths += [[0, 0]]  # 2-axis padding (no padding)
        array = np.pad(array, pad_widths, "reflect")

    if array.ndim == 2:
        sy, sx = array.shape
        nbands = False
    elif array.ndim == 3:
        sy, sx, nbands = array.shape
    else:
        raise ValueError(f"Incorrect array shape {array.shape}")

    # Calculate output shape
    if not nbands:
        strides = (
            array.strides[0] * step_y,
            array.strides[1] * step_x,
        ) + array.strides
        out_shape = (sy // step_y, sx // step_x, ksize_y, ksize_x)
    else:
        strides = (
            array.strides[0] * step_y,
            array.strides[1] * step_x,
            array.strides[2],
        ) + array.strides
        out_shape = (sy // step_y, sx // step_x, 1, ksize_y, ksize_x, nbands)

    # Slice the padded array using strides
    sliced_array = np.squeeze(as_strided(array, shape=out_shape, strides=strides))

    return sliced_array



import logging
import tempfile
import uuid
from numbers import Number
from pathlib import Path
from typing import Iterable, Union

import cv2
import geopandas as gpd
import numpy as np
from osgeo import gdal, gdalconst, ogr

from . import io
from .base import Image

logger = logging.getLogger(__name__)

__all__ = ["shapefile_masking"]


def shapefile_masking(
    gdf: gpd.GeoDataFrame,
    shape: Iterable[int],
    transformation: tuple,
    projection: str,
    mask_outpath: Union[Path, str] = None,
    burn_value: Number = 1,
    dilation: bool = False,
    dilation_iters: int = None,
    compression: bool = True,
) -> Image:
    """Rasterize polygons from a GeoDataFrame into a raster mask.

    :param gdf: Polygons as a :class:`geopandas.GeoDataFrame`.
    :param shape: Output array shape, e.g. ``(rows, cols)`` or ``(rows, cols, bands)``.
    :param transformation: GDAL geo-transform tuple/list.
    :param projection: CRS WKT string.
    :param mask_outpath: Optional path for the output mask GeoTIFF. If not provided, a
                         temporary file is used.
    :param burn_value: Value to burn into polygons.
    :param dilation: If ``True``, dilate the mask after rasterization.
    :param dilation_iters: Number of dilation iterations (defaults to 1 if
        ``dilation=True``).
    :param compression: If ``True``, enable GeoTIFF compression.
    :return: Mask as an :class:`~resens.base.Image`.
    """
    # Set up the output filename in a way that it won't be needed to create a
    # mask for arrays with the same extents
    remove_files = []
    out_epsg = gdf.crs.from_wkt(projection).to_epsg()
    if gdf.crs.to_epsg() != out_epsg:
        gdf = gdf.to_crs(epsg=out_epsg)
    polygon_shp = Path(tempfile.NamedTemporaryFile().name)
    gdf.to_file(polygon_shp)
    remove_files.append(polygon_shp)

    if mask_outpath:
        mask_outpath = Path(mask_outpath)
        if mask_outpath.suffix != ".tif":
            mask_outpath = mask_outpath.joinpath(f"land_mask_{str(uuid.uuid4())}.tif")
    else:
        mask_outpath = Path(tempfile.NamedTemporaryFile().name).with_suffix(".tif")
        remove_files.append(mask_outpath)
    mask_outpath.parent.mkdir(exist_ok=True, parents=True)

    # Write empty raster and load the dataset
    mask = Image(
        np.zeros(shape[:2], dtype=np.int8),
        transformation,
        projection,
    )
    mask.write_image(mask_outpath, compression)
    target_ds = gdal.Open(mask_outpath.as_posix(), gdalconst.GA_Update)

    # Load shapefile layers
    shp_ds = ogr.Open(polygon_shp.as_posix())
    shp_lyr = shp_ds.GetLayer()

    # Rasterization
    gdal.RasterizeLayer(
        target_ds, [1], shp_lyr, burn_values=[burn_value], options=["ALL_TOUCHED=TRUE"]
    )
    target_ds = None

    # Load the mask array
    mask = io.load_image(mask_outpath)

    # Dilate the mask
    if dilation:
        if dilation_iters:
            mask.array = cv2.dilate(
                mask.array,
                cv2.getStructuringElement(cv2.MORPH_ELLIPSE, ksize=(3, 3)),
                iterations=int(dilation_iters),
            )
        else:
            mask.array = cv2.dilate(
                mask.array,
                cv2.getStructuringElement(cv2.MORPH_ELLIPSE, ksize=(3, 3)),
                iterations=1,
            )
        mask.write_image(mask_outpath, compression=compression)

    for fil in remove_files:
        try:
            fil_path = Path(fil)
            if fil_path.is_file():
                fil_path.unlink()
            elif fil_path.is_dir():
                for child in fil_path.iterdir():
                    child.unlink()
                fil_path.rmdir()
        except Exception:
            # Best-effort cleanup; ignore errors
            pass

    return mask
